package git;

import issueData.Issue;
import org.eclipse.jgit.api.*;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.dircache.DirCache;
import org.eclipse.jgit.lib.*;
import org.eclipse.jgit.merge.MergeStrategy;
import org.eclipse.jgit.merge.Merger;
import org.eclipse.jgit.merge.ResolveMerger;
import org.eclipse.jgit.merge.ThreeWayMerger;
import org.eclipse.jgit.revwalk.RevCommit;
import org.eclipse.jgit.revwalk.RevTree;
import org.eclipse.jgit.revwalk.RevWalk;
import org.eclipse.jgit.revwalk.filter.RevFilter;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;
import org.eclipse.jgit.transport.RefSpec;
import org.eclipse.jgit.treewalk.AbstractTreeIterator;
import org.eclipse.jgit.treewalk.TreeWalk;

import java.awt.*;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Scanner;

public class IssueRepo {

    private Repository repository;
    private ObjectInserter oi;

    public IssueRepo() {
        try {
            this.repository = new FileRepositoryBuilder()
                    .setGitDir(new File("/home/jack/Documents/testrepo/.git"))
//                    .setGitDir(new File("/home/jack/Documents/testrepo2/.git"))
                    .build();

        } catch (IOException e) {
            e.printStackTrace();
        }
        oi = new ObjectInserter(repository);
    }

    public void push(String remote) {
        try (Git git = new Git(repository)) {
            try {
                git.push().setRemote(remote).setForce(true).setRefSpecs(new RefSpec("refs/heads/issues:refs/heads/issues")).call();
            } catch (GitAPIException e) {
                e.printStackTrace();
            }
        }
    }

    public void pull(String remote) {
//        localClone();
        pullOnClone(remote);
//

    }

    private void anotherMerge(RevCommit commit) {
        try (Git git = new Git(repository)) {
//            StoredConfig config = git.getRepository().getConfig();
//            config.setString("remote", "remIssues", "url", "/home/jack/Documents/clone/");
//            config.save();
//
//            git.fetch().setRemote("remIssues")
//                    .setRefSpecs(new RefSpec("refs/heads/issues:refs/remotes/remIssues"))
//                    .call();

            Ref localRef = repository.findRef("refs/heads/issues");
//            Ref remoteRef = repository.findRef("refs/remotes/remIssues");

            merge(localRef.getObjectId(), commit);

        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void localClone() {
        try {
            Git git = Git.cloneRepository()
                    .setURI("/home/jack/Documents/testrepo/.git")
                    .setDirectory(new File("/home/jack/Documents/clone"))
                    .setBranchesToClone(Collections.singleton("refs/heads/issues"))
                    .setBranch("refs/heads/issues")
                    .call();
        } catch (GitAPIException e) {
            e.printStackTrace();
        }
    }

    private void pullOnClone(String remote) {
        try (Git git = Git.open(new File("/home/jack/Documents/clone"))) {

            StoredConfig config = git.getRepository().getConfig();
            config.setString("remote", "remIssues", "url", remote);
            config.save();

            git.fetch()
                    .setRemote("remIssues")
                    .setRefSpecs(new RefSpec("refs/heads/issues:refs/remotes/remIssues"))
                    .call();

            ObjectId mergeBase = git.getRepository().resolve("remIssues");

            MergeCommand merge = git.merge()
                    .include(mergeBase)
                    .setCommit(false)
                    .setMessage("Merged");
            MergeResult mergeResult = merge.call();

            if (!mergeResult.getMergeStatus().isSuccessful() && mergeResult.getConflicts() != null) {
                System.out.println("Conflicts exist in:");
                mergeResult.getConflicts().forEach((k, v) -> System.out.println(k));

                if (Desktop.isDesktopSupported()) {
                    List<Thread> threads = new ArrayList<>();

                    mergeResult.getConflicts().forEach((k, v) -> {
                        Runnable runnable = () -> {
                            try {
                                Desktop.getDesktop().open(new File("/home/jack/Documents/clone/" + k));
                            } catch (IOException e) {
                                e.printStackTrace();
                            }
                        };
                        Thread thread = new Thread(runnable);
                        threads.add(thread);
                    });

                    threads.forEach(Thread::start);

                    threads.forEach(t -> {
                        try {
                            t.join();
                        } catch (InterruptedException e) {
                            e.printStackTrace();
                        }
                    });

                    System.out.println("Enter \"abort\" to stop merge or \"resolved\" if finished resolving - any other input will be treated as abort");
                    Scanner scanner = new Scanner(System.in);
                    String input = scanner.nextLine();
                    scanner.close();

                    if (input.equals("resolved")) {
                        git.add().addFilepattern(".").call();
                        RevCommit commit = git.commit().setMessage("resolved merge").call();
                        anotherMerge(commit);
                        System.out.println("maybe solved?");
                    } else {
                        git.reset().setMode(ResetCommand.ResetType.HARD).call();
                        System.out.println("Merge abandoned");
                    }

                }

            }

        } catch (IOException | GitAPIException e) {
            e.printStackTrace();
        }
    }

    private void merge(ObjectId localRef, RevCommit remoteRef) {

        RevCommit localCommit = null;
        RevCommit remoteCommit = null;
        RevCommit mergeBase = null;

        try (RevWalk revWalk = new RevWalk(repository)) {
            localCommit = revWalk.parseCommit(localRef);
//            remoteCommit = revWalk.parseCommit(remoteRef);

//            revWalk.setRevFilter(RevFilter.MERGE_BASE);
//            revWalk.markStart(localCommit);
//            revWalk.markStart(remoteCommit);
//            mergeBase = revWalk.next();

            revWalk.dispose();
        } catch (IOException e) {
            e.printStackTrace();
        }

        Merger merger;
        MergeStrategy strategy;

//        if (mergeBase == null) {
            strategy = MergeStrategy.THEIRS;
            System.out.println("theirs");
            merger = strategy.newMerger(repository);
//        } else {
//            strategy = MergeStrategy.RECURSIVE;
//            System.out.println("recursive");
//            merger = strategy.newMerger(repository);
//            try {
//                ((ThreeWayMerger) merger).setBase(mergeBase);
//            } catch (IOException e) {
//                e.printStackTrace();
//            }
//        }

        try {
            if (merger.merge(localCommit, remoteRef)) {
                String commit = "Merged " + localRef.getName() + " and " + remoteRef.getName();
                commit(commit, merger.getResultTreeId(), localCommit, remoteRef);
            } else {
                System.out.println(((ResolveMerger) merger).getFailingPaths());
                for (ResolveMerger.MergeFailureReason c : ResolveMerger.MergeFailureReason.values())
                    System.out.println(c);
                System.out.println(((ResolveMerger) merger).getModifiedFiles());
                System.out.println("unmerged");
            }
        } catch (IOException e) {
            e.printStackTrace();
        }

    }


    public void writeIssue(Issue issue, String commitMessage) throws IOException {
//        DirCache index = repository.lockDirCache();

        RevCommit currentCommit = getCurrentCommit();

        IssueTreeBuilder issueWriter = new IssueTreeBuilder(repository);
        ObjectId issueTree = issueWriter.buildIssueTree(issue);

        TreeFormatter repoTreeFormatter = getTree(currentCommit.getTree());
        repoTreeFormatter.append(issue.getOriginalHash(), FileMode.TREE, issueTree);

        ObjectId repoTree = oi.insert(Constants.OBJ_TREE, repoTreeFormatter.toByteArray());

        commit(commitMessage, repoTree, currentCommit);
//        index.unlock();
    }

    public void updateIssue(Issue issue, String commitMessage) throws IOException {
        RevCommit currentCommit = getCurrentCommit();

        IssueTreeBuilder issueWriter = new IssueTreeBuilder(repository);
        ObjectId issueTree = issueWriter.buildIssueTree(issue);

        TreeFormatter repoTreeFormatter = removeFromRepoTree(currentCommit.getTree(), issue.getOriginalHash());
        repoTreeFormatter.append(issue.getOriginalHash(), FileMode.TREE, issueTree);

        ObjectId repoTree = oi.insert(Constants.OBJ_TREE, repoTreeFormatter.toByteArray());

        commit(commitMessage, repoTree, currentCommit);
    }

    public void deleteIssue(String hash) throws IOException {
        DirCache index = repository.lockDirCache();

        RevCommit currentCommit = getCurrentCommit();
        TreeFormatter repoTreeFormatter = removeFromRepoTree(currentCommit.getTree(), hash);

        ObjectId repoTree = oi.insert(Constants.OBJ_TREE, repoTreeFormatter.toByteArray());

        commit("Deleted issue " + hash, repoTree, currentCommit);

        index.unlock();

    }

    private TreeFormatter removeFromRepoTree(RevTree revTree, String hash) throws IOException {
        TreeFormatter issuesTreeFormatter = new TreeFormatter();
        try (TreeWalk treeWalk = new TreeWalk(repository)) {
            treeWalk.addTree(revTree);
            treeWalk.setRecursive(true);
            while (treeWalk.next()) {
                if (treeWalk.getTreeCount() > 0) {
                    AbstractTreeIterator iterator = treeWalk.getTree(0, AbstractTreeIterator.class);
                    if (!iterator.getEntryPathString().contains(hash)) {
                        issuesTreeFormatter.append(iterator.getEntryPathBuffer(), 0,
                                iterator.getEntryPathLength(), iterator.getEntryFileMode(),
                                iterator.getEntryObjectId());
                    }
                }
            }
        }

        return issuesTreeFormatter;
    }

    private TreeFormatter getTree(RevTree revTree) throws IOException {
        TreeFormatter issuesTreeFormatter = new TreeFormatter();

        try (TreeWalk treeWalk = new TreeWalk(repository)) {
            treeWalk.addTree(revTree);
            treeWalk.setRecursive(true);
            while (treeWalk.next()) {
                if (treeWalk.getTreeCount() > 0) {
                    AbstractTreeIterator iterator = treeWalk.getTree(0, AbstractTreeIterator.class);
                    issuesTreeFormatter.append(iterator.getEntryPathBuffer(), 0,
                            iterator.getEntryPathLength(), iterator.getEntryFileMode(),
                            iterator.getEntryObjectId());
                }
            }
        }

        return issuesTreeFormatter;
    }

    public List<Issue> getIssues() {
        try {
            IssueReader issueReader = new IssueReader(repository, getCurrentCommit().getTree());
            return issueReader.getIssues();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return new ArrayList<>();
    }

    private void commit(String message, ObjectId tree, ObjectId... parent) throws IOException {
        CommitBuilder commitBuilder = new CommitBuilder();
        commitBuilder.setMessage(message);
        PersonIdent personIdent = new PersonIdent(getUserName(), getUserEmail());
        commitBuilder.setAuthor(personIdent);
        commitBuilder.setCommitter(personIdent);
        if (tree != null && parent != null) {
            commitBuilder.setTreeId(tree);
            commitBuilder.setParentIds(parent);
        } else {
            commitBuilder.setTreeId(oi.insert(Constants.OBJ_TREE, new byte[]{}));
        }

        //update issues branch ref to point to new commit
        ObjectId newCommit = oi.insert(commitBuilder);
        RefUpdate update = repository.updateRef("refs/heads/issues");
        update.setNewObjectId(newCommit);
        update.setForceUpdate(true);
        update.update();
    }

    public void commit(String message) throws IOException {
        commit(message, null);
    }

    private RevCommit getCurrentCommit() throws IOException {
        RevCommit commit;
        Ref issueRef = repository.findRef("refs/heads/issues");

        try (RevWalk walk = new RevWalk(repository)) {
            commit = walk.parseCommit(issueRef.getObjectId());
            walk.dispose();
            return commit;
        }
    }

    private String getUserName() {
        String name = repository.getConfig().getString("user", null, "name");
        if (name == null) {
            return "Username unknown ";
        } else {
            return "User " + name;
        }
    }

    private String getUserEmail() {
        String email = repository.getConfig().getString("user", null, "email");
        if (email == null) {
            return "Email unknown";
        } else {
            return "Email " + email;
        }
    }

    public String getUserAndEmail() {
        return getUserName() + " " + getUserEmail();
    }

}
