package git;

import issueData.Comment;
import issueData.Issue;
import org.eclipse.jgit.lib.*;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

public class IssueTreeBuilder {

    private ObjectInserter oi;

    public IssueTreeBuilder(Repository repository) {
        this.oi = new ObjectInserter(repository);
    }

    public ObjectId buildIssueTree(Issue issue) throws IOException {
        TreeFormatter issueTreeFormatter = new TreeFormatter();

        ObjectId title = oi.insert(Constants.OBJ_BLOB, issue.getTitle().getBytes());
        ObjectId description = oi.insert(Constants.OBJ_BLOB, issue.getDescription().getBytes());
        ObjectId creator = oi.insert(Constants.OBJ_BLOB, issue.getCreator().getBytes());
        ObjectId status = oi.insert(Constants.OBJ_BLOB, String.valueOf(issue.isopen()).getBytes());
        ObjectId creationTime = oi.insert(Constants.OBJ_BLOB, String.valueOf(issue.getCreationTime()).getBytes());
        ObjectId editTime = oi.insert(Constants.OBJ_BLOB, String.valueOf(issue.getEdited()).getBytes());
        ObjectId tags = oi.insert(Constants.OBJ_BLOB, buildTagString(issue.getTags()).getBytes());
        ObjectId comments = buildCommentTree(issue.getComments());

        issueTreeFormatter.append("title", FileMode.REGULAR_FILE, title);
        issueTreeFormatter.append("description", FileMode.REGULAR_FILE, description);
        issueTreeFormatter.append("creator", FileMode.REGULAR_FILE, creator);
        issueTreeFormatter.append("status", FileMode.REGULAR_FILE, status);
        issueTreeFormatter.append("creationTime", FileMode.REGULAR_FILE, creationTime);
        issueTreeFormatter.append("editTime", FileMode.REGULAR_FILE, editTime);
        issueTreeFormatter.append("tags", FileMode.REGULAR_FILE, tags);
        issueTreeFormatter.append("comments", FileMode.TREE, comments);


        return oi.insert(Constants.OBJ_TREE, issueTreeFormatter.toByteArray());
    }

    private ObjectId buildCommentTree(List<Comment> comments) throws IOException {
        TreeFormatter allCommentTreeFormatter = new TreeFormatter();
        for (Comment comment : comments) {

            TreeFormatter commentTreeFormatter = new TreeFormatter();
            ObjectId author = oi.insert(Constants.OBJ_BLOB, comment.getAuthor().getBytes());
            ObjectId message = oi.insert(Constants.OBJ_BLOB, comment.getMessage().getBytes());
            ObjectId commentTime = oi.insert(Constants.OBJ_BLOB, String.valueOf(comment.getCreationTime()).getBytes());

            commentTreeFormatter.append("author", FileMode.REGULAR_FILE, author);
            commentTreeFormatter.append("message", FileMode.REGULAR_FILE, message);
            commentTreeFormatter.append("commentTime", FileMode.REGULAR_FILE, commentTime);

            ObjectId commentTree = oi.insert(Constants.OBJ_TREE, commentTreeFormatter.toByteArray());

            allCommentTreeFormatter.append(comment.getHash(), FileMode.TREE, commentTree);
        }

        return oi.insert(Constants.OBJ_TREE, allCommentTreeFormatter.toByteArray());
    }

    private String buildTagString(Set<String> tags) {
        StringBuilder stringBuilder = new StringBuilder();
        Iterator<String> iterator = tags.iterator();

        if (!iterator.hasNext()) {
            return "";
        }

        for (; ; ) {
            stringBuilder.append(iterator.next());
            if (!iterator.hasNext()) {
                return stringBuilder.toString();
            }
            stringBuilder.append(", ");
        }
    }


//
//    public ObjectId buildIssueTree(Issue issue) throws IOException {
//        TreeFormatter issueTreeFormatter = new TreeFormatter();
//        String path = "/home/jack/Documents/tmp/" + issue.getOriginalHash() + "/";
//
//        ObjectId title = insert(Constants.OBJ_BLOB, Files.readAllBytes(Paths.get(path + "title")));
//        ObjectId description = insert(Constants.OBJ_BLOB, Files.readAllBytes(Paths.get(path + "description")));
//        ObjectId creator = insert(Constants.OBJ_BLOB, Files.readAllBytes(Paths.get(path + "creator")));
//        ObjectId status = insert(Constants.OBJ_BLOB, Files.readAllBytes(Paths.get(path + "status")));
//        ObjectId creationTime = insert(Constants.OBJ_BLOB, Files.readAllBytes(Paths.get(path + "creationTime")));
//        ObjectId editTime = insert(Constants.OBJ_BLOB, Files.readAllBytes(Paths.get(path + "editedTime")));
//        ObjectId comments = buildCommentTree(issue.getComments());
//
//        issueTreeFormatter.append("title", FileMode.REGULAR_FILE, title);
//        issueTreeFormatter.append("description", FileMode.REGULAR_FILE, description);
//        issueTreeFormatter.append("creator", FileMode.REGULAR_FILE, creator);
//        issueTreeFormatter.append("status", FileMode.REGULAR_FILE, status);
//        issueTreeFormatter.append("creationTime", FileMode.REGULAR_FILE, creationTime);
//        issueTreeFormatter.append("editTime", FileMode.REGULAR_FILE, editTime);
//        issueTreeFormatter.append("comments", FileMode.TREE, comments);
//
//        return insert(Constants.OBJ_TREE, issueTreeFormatter.toByteArray());
//    }


}
